define
{
   LEX_ERR = 0

   OPER_RPARENTH      // )
   OPER_LPARENTH      // (
   OPER_PLUS          // +
   OPER_MINUS         // -
   OPER_MUL           // *
   OPER_DIV           // /
   OPER_AND           // &
   OPER_OR            // |
   OPER_LSHIFT        // <<
   OPER_RSHIFT        // >>
   OPER_UPLUS         // unary +
   OPER_UMINUS        // unary -
   OPER_UNOT          // unary ~
   
   NUM_LONG           // long value
   NUM_DOUBLE         // double value
   NUM_A NUM_B NUM_C NUM_D NUM_E NUM_F NUM_G   // registers
      
   LEX_NONE
}

type lexem
{
   uint   id
   long   lval
   double dval
}

global
{
   arr  oper     = %{ 0, ')', '(', '+', '-', '*', '/', '&', '|', 0x3C3C, 0x3E3E, '+', '-', '~' }
   arr  priority = %{ 0,   0,   1,   2,   2,   3,   3,   3,   3,      3,      3,   4,   4,   4 }
   arr  register[7] of lexem
}

func uint getlexem( str expr, lexem lex )
{
   uint ch ptr result = $LEX_ERR
   str  stemp
   
   expr.trimspace()
   
   ptr = expr.ptr()
   if !( ch = ptr->byte )
   {
      result = $LEX_NONE
      goto end       
   }
   switch ch 
   {
      case '+', '-', '*', '/', '(', ')', '&', '|', '~' 
      {
         uint i
         fornum i, *oper
         {
            if oper[i] == ch
            { 
               result = i
               break
            }
         }
         ptr++
      }
      case '<', '>' 
      {
         ptr++
         if ptr->byte != ch : goto end
         ptr++
         result = ?( ch == '<', $OPER_LSHIFT, $OPER_RSHIFT )
      }
      default
      {
         if ch >= 'A' && ch <= 'G'
         {
            ptr++
            result = $NUM_A + ch - 'A'
            goto end
         } 
         if ch < '0' || ch > '9' : goto end

         result = $NUM_LONG
         if "0X".eqlen( expr )
         {
            ptr += 2
            while ( ptr->byte >= '0' && ptr->byte <= '9' ) ||
                  ( ptr->byte >= 'A' && ptr->byte <= 'F' ) : ptr++
         }
         while  ptr->byte >= '0' && ptr->byte <= '9'
         {
            ptr++
            if ptr->byte == '.'
            {
               if result == $NUM_DOUBLE : goto end
               result = $NUM_DOUBLE
               ptr++     
            }
            elif ptr->byte == 'E'
            {
               ptr++
               result = $NUM_DOUBLE
               if ptr->byte == '+' || ptr->byte == '-' : ptr++
               while  ptr->byte >= '0' && ptr->byte <= '9' : ptr++
            }
         }
         stemp.copy( expr.ptr(), ptr - expr.ptr() ) 
         if result == $NUM_LONG : lex.lval = str2long( stemp )
         elif result == $NUM_DOUBLE : lex.dval = str2double( stemp )
      }
   }
   
   label end
   expr.del( 0, ptr - expr.ptr())
   lex.id = result
    
   return result
} 

operator lexem =( lexem left right )
{
   left.id = right.id
   left.lval = right.lval
   left.dval = right.dval
   return left
}

func uint calculate( str expr, lexem lex )
{
   stack  vals of lexem
   stack  opers of lexem
   str    expression 
   uint   prevlex top parenth

   subfunc uint calc
   {
      lexem curoper
      uint  left

      while opers.top()
      {
         lexem right

         if !*vals : return 0
         curoper = opers.top()->lexem
         if lex.id == $OPER_RPARENTH
         {
            if curoper.id == $OPER_LPARENTH
            {
               opers.pop() 
               break
            }
         }
         elif priority[ curoper.id ] < priority[ lex.id ] : break
         if curoper.id < $OPER_UPLUS
         {
            right = vals.top()->lexem
            vals.pop()
            if !*vals : return 0
         }
         left as vals.top()->lexem
         if curoper.id < $OPER_AND && left.id != right.id
         {
            if left.id == $NUM_LONG
            { 
               left.dval = double( left.lval )
               left.id = $NUM_DOUBLE
            }   
            if right.id == $NUM_LONG
            { 
               right.dval = double( right.lval )
               right.id = $NUM_DOUBLE
            }   
         }
         if left.id == $NUM_DOUBLE && ( curoper.id == $OPER_UNOT ||
            ( curoper.id >= $OPER_AND && curoper.id >= $OPER_RSHIFT )) : return 0
         switch curoper.id
         {
            case $OPER_PLUS 
            {
               if left.id == $NUM_LONG : left.lval += right.lval
               else : left.dval += right.dval              
            }
            case $OPER_MINUS 
            {
               if left.id == $NUM_LONG : left.lval -= right.lval
               else : left.dval -= right.dval              
            }
            case $OPER_MUL 
            {
               if left.id == $NUM_LONG : left.lval *= right.lval
               else : left.dval *= right.dval              
            }
            case $OPER_DIV 
            {
               if left.id == $NUM_LONG
               { 
                  if !right.lval : return 0 
                  left.lval /= right.lval
               }
               else : left.dval /= right.dval              
            }
            case $OPER_AND : left.lval &= right.lval
            case $OPER_OR : left.lval |= right.lval
            case $OPER_LSHIFT : left.lval <<= right.lval
            case $OPER_RSHIFT : left.lval >>= right.lval
            case $OPER_UMINUS
            {
               if left.id == $NUM_LONG : left.lval = -left.lval
               else : left.dval = -left.dval              
            }  
            case $OPER_UNOT : left.lval = ~left.lval
         }
         opers.pop()         
      }
      return 1   
   }
   
   ( expression = expr ).upper()
   do
   {
      getlexem( expression, lex )
//      print(" ID= \(lex.id) \n ")

      if lex.id == $LEX_NONE : break
      elif lex.id == $LEX_ERR : return $LEX_ERR
      elif lex.id >= $NUM_LONG
      { 
         if prevlex >= $NUM_LONG || prevlex == $OPER_RPARENTH : return $LEX_ERR
         if lex.id >= $NUM_A : lex = register[ lex.id - $NUM_A ]
         vals.push()->lexem = lex
      }
      else
      {
         if prevlex < $NUM_LONG && prevlex != $OPER_RPARENTH  
         {
            switch lex.id
            { 
               case $OPER_PLUS : lex.id = $OPER_UPLUS
               case $OPER_MINUS : lex.id = $OPER_UMINUS
               default
               { 
                  if lex.id != $OPER_UNOT && lex.id != $OPER_LPARENTH
                  { 
                     return $LEX_ERR
                  }
               }
            }   
         }
         if lex.id == $OPER_RPARENTH
         { 
            if !parenth || !calc() : return $LEX_ERR
            parenth--
         }
         else
         {                                              
            top as opers.top()->lexem
            if ( top && priority[ lex.id ] <= priority[ top.id ] && 
                 priority[ lex.id ] != 4 && lex.id !=  $OPER_LPARENTH )
            { 
               if !calc() : return $LEX_ERR
            }
            opers.push()->lexem = lex     
         }
         if lex.id == $OPER_LPARENTH : parenth++                                              
      }
      prevlex = lex.id
   } while *expression
   
   lex.id = 0
   
   if !calc() : return $LEX_ERR
   if vals.top() : lex = vals.top()->lexem   
   
   return lex.id   
}

/*func main<main>
{
   str   expr
   lexem result
   uint  i
   
   fornum i, 7
   {
      register[i].id = $NUM_LONG
      register[i].lval = long( i )
   }
   do 
   {
      congetstr( "Enter a mathematical expression: ", expr )
      switch ( calculate( expr, result ))
      {
         case $LEX_ERR : @"There is an error in the expression\n" 
         case $NUM_LONG : @"Long = \( result.lval )\n"         
         case $NUM_DOUBLE : @"Double = \( result.dval )\n"         
      }      
	} while *expr
}*/
